Per-fragment Blinn-Phong shader for a single spot light source.

Specular reflections are calculated using a local viewer lighting model
(i.e., as if GL_LIGHT_MODEL_LOCAL_VIEWER is set to GL_TRUE).

We borrow the Direct3D convention of an inner and an outer cone for the spot
light. A direct per-fragment implementation of the OpenGL spot light model will
produce a very sharp falloff around the spotCutoff. Using the inner and outer
cone convention will produce a more softer falloff around the spotCutoff.

To enable the inner and outer cone version of the spot light fragment shader
define the preprocessor symbol USE_INNER_AND_OUTER_CONES. If this symbol is not
defined the fragment shader will use the OpenGL spot light model.

[vert]

varying vec3 position;
varying vec3 normal;

void main()
{
    vec4 eyeCoordPos = gl_ModelViewMatrix * gl_Vertex;

    position = eyeCoordPos.xyz / eyeCoordPos.w;
    normal = normalize(gl_NormalMatrix * gl_Normal);

    gl_Position = gl_ModelViewProjectionMatrix * gl_Vertex;
    gl_TexCoord[0] = gl_MultiTexCoord0;    
}

[frag]

#define USE_INNER_AND_OUTER_CONES

uniform sampler2D colorMap;

#if defined(USE_INNER_AND_OUTER_CONES)
    uniform float cosOuterCone;
    uniform float cosInnerCone;
#endif

varying vec3 position;
varying vec3 normal;

void main()
{   
    vec3 l = gl_LightSource[0].position.xyz - position;
    float d = length(l);
    float atten = 1.0 / (gl_LightSource[0].constantAttenuation +
                         gl_LightSource[0].linearAttenuation * d +
                         gl_LightSource[0].quadraticAttenuation * d * d);
    
    l = normalize(l);

    float spotDot = dot(-l, normalize(gl_LightSource[0].spotDirection));

#if defined(USE_INNER_AND_OUTER_CONES)
    float spotEffect = smoothstep(cosOuterCone, cosInnerCone, spotDot);
#else    
    float spotEffect = (spotDot < gl_LightSource[0].spotCosCutoff)
                        ? 0.0 : pow(spotDot, gl_LightSource[0].spotExponent);
#endif

    atten *= spotEffect;

    vec3 n = normalize(normal);
    vec3 v = normalize(-position);
    vec3 h = normalize(l + v);
    
    float nDotL = max(0.0, dot(n, l));
    float nDotH = max(0.0, dot(n, h));
    float power = (nDotL == 0.0) ? 0.0 : pow(nDotH, gl_FrontMaterial.shininess);
    
    vec4 ambient = gl_FrontLightProduct[0].ambient * atten;
    vec4 diffuse = gl_FrontLightProduct[0].diffuse * nDotL * atten;
    vec4 specular = gl_FrontLightProduct[0].specular * power * atten;
    vec4 color = gl_FrontLightModelProduct.sceneColor + ambient + diffuse + specular;
    
    gl_FragColor = color * texture2D(colorMap, gl_TexCoord[0].st);
}
